"""
Module to handle Matlab files.
"""
import os
import numpy as np
import scipy.io as sio


class MatlabHandler(object):
    """
    Matlab format file handler class.
    You are NOT supposed to call directly this class constructor (use
    :class:`.FileHandler` constructor instead).

    :cvar str _filename: name of file to handle.
    :cvar dict _cached_data: private attribute to save a copy of last data
        processed; it allows to reduce IO operation on the same file. Initial
        value is None.

    """

    def __init__(self, filename):

        self._filename = filename
        self._cached_data = None

    def get_dataset(self, output_name):
        """
        Method to parse the `filename`. It returns a matrix with all the values
        of the chosen output.

        :param str output_name: name of the output of interest inside the
            mat file.

        :return: a *n_points*-by-*n_components* matrix containing the values of
            the chosen output.
        :rtype: numpy.ndarray
        """
        if self._cached_data is None:
            if not os.path.isfile(self._filename):
                raise RuntimeError("{0!s} doesn't exist".format(
                    os.path.abspath(self._filename)))

            self._cached_data = sio.loadmat(self._filename)

        if not output_name in self._cached_data:
            raise RuntimeError("File has no " + output_name + " field.")

        return np.array(self._cached_data[output_name])

    def set_dataset(self, output_values, output_name):
        """
        Writes to filename the given output. `output_values` is a matrix that
        contains the new values of output to write, `output_name` is a string
        that indicates name of output to write.

        :param numpy.ndarray output_values: it is a
            *n_points*-by-*n_components* matrix containing the output values.
        :param str output_name: name of the output.
        """
        sio.savemat(self._filename, {output_name: output_values})
        self._cached_data = None



__all__ = [
    'filehandler', 'matlabhandler', 'vtkhandler', 'podinterpolation', 'online',
    'stlhandler', 'mapper', 'offline', 'utilities', 'points', 'snapshots',
    'parametricspace', 'interpolation', 'rbf'
]

from . import filehandler
from . import matlabhandler
from . import interpolation
from . import vtkhandler
from . import stlhandler
from . import podinterpolation
from . import parametricspace
from . import online
from . import offline
from . import mapper
from . import points
from . import snapshots
from . import utilities
from .ndinterpolator import rbf



"""
Module with the base class for a generic space.
"""
import pickle


class ParametricSpace(object):
    """
    Abstract class.
    """

    def __init__(self):
        raise NotImplementedError

    def __call__(self, value):
        """
        Abstract method to approximate the value of a generic point.

        Not implemented, it has to be implemented in subclasses.
        """
        raise NotImplementedError

    def save(self, filename):
        """
        Save the space on file.

        :param str filename: the filename where the space has to be saved.
        """
        with open(filename, 'w') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(filename):
        """
        Load the space from a file.

        :param str filename: the filename where the space has been saved.
        :return: the space instance.
        :rtype: ParametricSpace
        """
        with open(filename, 'r') as f:
            return pickle.load(f)



"""
Module for Radial Basis Function Interpolation
"""

import numpy as np
from scipy.spatial.distance import cdist as distance_matrix


class RBFInterpolator(object):
    """
    Multidimensional interpolator using Radial Basis Function.

    :param array_like points: the coordinates of the points.
    :param array_like values: the values in the points.
    :param float radius: the radius used in the basis functions. Default is 1.0.
    :param norm: The function that returns the distance between two points. It
        has to be a function that take as input a vector and return a float
        number.  Default is 'euclidean'.
    :type norm: str or callable
    :param callable basis: the basis function.

    :cvar array_like points: the coordinates of the points.
    :cvar float radius: the radius used in the basis functions. Default is 1.0.
    :cvar callable basis: the basis function.
    :cvar numpy.ndarray weights: the weights matrix.
    """

    @staticmethod
    def multi_quadratic(X, r):
        """
        It implements the following formula:

        .. math::

            \\varphi( \\boldsymbol{x} ) = \\sqrt{\\boldsymbol{x}^2 +
            r^2}

        :param X: the vector x in the formula above.
        :type X: array_like or float
        :param float r: the parameter r in the formula above.
        :return: the result of the formula above.
        :rtype: array_like or float
        """
        return np.sqrt(r**2 + X**2)

    def __init__(self, points, values, radius=1.0, norm='euclidean',
                 basis=None):

        self.basis = basis
        self.points = points
        self.radius = radius

        if self.basis is None:
            self.basis = self.multi_quadratic

        self.weights = np.linalg.solve(
            self.basis(
                distance_matrix(points, points, metric=norm), self.radius),
            values)

    def __call__(self, new_points):
        """
        Evaluate interpolator at given `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        return self.basis(
            distance_matrix(new_points, self.points), self.radius).dot(
                self.weights)



"""
Class for computation of the Offline part. It provides methods for:
    - import the `snapshots` and the parameter values
      correlated(:func:`init_database <ezyrb.offline.Offline.init_database>` ,
      :func:`init_database_from_file
      <ezyrb.offline.Offline.init_database_from_file>`);
    - generate the reduced space with the method chosen by the user;
    - estimate the error using the `leave-one-out` strategy;
    - save the reduced space to a specific file.
"""

import os
import numpy as np
from ezyrb.podinterpolation import PODInterpolation
from ezyrb.points import Points
from ezyrb.snapshots import Snapshots
from ezyrb.utilities import simplex_volume


class Offline(object):
    """
    Offline phase.

    :param ParametricSpace spacetype: the method used for the reduced space
        generation. Default is :class:`.PODInterpolation`.
    :param str output_name: the name of the output of interest.
    :param str weight_name: the name of the output to consider as weight.
    :param str dformat: the data format to extract from the snapshot files:
        if the parameter is "cell", the snapshot values refer to the cell data,
        if the parameter is "point", the snapshot values refer to the point
        data. These are the only options available.

    :cvar Points mu_values: the object that contains the parameter values
    :cvar Snapshots snapshots: the object that contains the snapshots extracted
        from the chosen files.
    :cvar ParametricSpace spacetype: the method used for
        the reduced space generation.
    """

    def __init__(self,
                 output_name,
                 space_type=PODInterpolation,
                 weight_name=None,
                 dformat='cell'):

        self.mu = Points()
        self.snapshots = Snapshots(output_name, weight_name, dformat)
        self.space = space_type()

    def init_database(self, values, files):
        """
        Initialize the database with the passed parameter values and snapshot
        files: the *i*-th parameter has to be the parametric point of the
        solution stored in the *i*-th file.

        :param array_like values: the list of parameter values.
        :param array_like files: the list of the solution files.
        """
        for mu in values:
            self.mu.append(mu)
        for fl in files:
            self.snapshots.append(fl)

    def init_database_from_file(self, filename):
        """
        Initialize the database by reading the parameter values and the
        snapshot files from a given filename; this file has to be format as
        following: first *N* columns indicate the parameter values, the *N+1*
        column indicates the corresponding snapshot file.

        Example of a generic file::

            par1    par2    ...     solution_file1
            par1    par2    ...     solution_file2
            par1    par2    ...     solution_file3

        :param str filename: name of file where parameter values and snapshot
            files are stored.
        """
        if not os.path.isfile(filename):
            raise IOError("File {0!s} not found".format(
                os.path.abspath(filename)))

        matrix = np.genfromtxt(filename, dtype=None)
        num_cols = len(matrix.dtype)

        if num_cols < 2:
            raise ValueError("Not valid number of columns in database file")

        snapshots_files = matrix[matrix.dtype.names[num_cols - 1]].astype(str)
        mu_values = np.array([
            matrix[name].astype(float)
            for name in matrix.dtype.names[0:num_cols - 1]
        ])

        self.init_database(mu_values.T, snapshots_files)

    def add_snapshot(self, new_mu, new_file):
        """
        This methos adds the new solution to the database and the new parameter
        values to the parameter points; this can be done only after the new
        solution has be computed and placed in the proper directory.

        :param array_like new_mu: the parameter value to add to database.
        :param str new_file: the name of snapshot file to add to
            database.
        """
        self.mu.append(new_mu)
        self.snapshots.append(new_file)

    def generate_rb_space(self):
        """
        Generate the reduced basis space by combining the snapshots. It
        uses the chosen method for the generation.
        """
        self.space.generate(self.mu, self.snapshots)

    def save_rb_space(self, filename):
        """
        Save the reduced basis space to `filename`.

        :param str filename: the file where the space will be stored.
        """
        self.space.save(filename)

    def loo_error(self, func=np.linalg.norm):
        """
        Estimate the approximation error using *leave-one-out* strategy. The
        main idea is to create several reduced spaces by combining all the
        snapshots except one. The error vector is computed as the difference
        between the removed snapshot and the projection onto the properly
        reduced space. The procedure repeats for each snapshot in the database.
        The `func` is applied on each vector of error to obtained a float
        number.

        :param function func: the function used to assign at each vector of
            error a float number. It has to take as input a 'numpy.ndarray` and
            returns a float. Default value is the L2 norm.
        :return: the vector that contains the errors estimated for all
            parametric points.
        :rtype: numpy.ndarray
        """
        return self.space.loo_error(self.mu, self.snapshots, func)

    def optimal_mu(self, error=None, k=1):
        """
        Return the parametric points where new high-fidelity solutions have to
        be computed in ordere to globaly reduce the estimated error. These
        points are the barycentric center of the region (simplex) with higher
        error.

        :param numpy.ndarray error: the estimated error evaluated for each
            snapshot; if error array is not passed, it is computed using
            :func:`loo_error` with the default function. Default value is None.
        :param int k: the number of optimal points to return. Default value is
            1.
        :return: the optimal points
        :rtype: list(numpy.ndarray)
        """
        if error is None:
            error = self.loo_error()

        tria = self.mu.triangulation

        error_on_simplex = np.array([
            np.sum(error[smpx]) * simplex_volume(self.mu.values.T[smpx])
            for smpx in tria.simplices
        ])

        barycentric_point = []
        for index in np.argpartition(error_on_simplex, -k)[-k:]:
            worst_tria_pts = self.mu.values.T[tria.simplices[index]]
            worst_tria_err = error[tria.simplices[index]]

            barycentric_point.append(
                np.average(worst_tria_pts, axis=0, weights=worst_tria_err))

        return barycentric_point



"""
Module for the Points class.

It provides an easy interface to collect parametric points.
"""
import numpy as np
from scipy.spatial import Delaunay


class Points(object):
    """
    :cvar numpy.ndarray _values: the matrix that contains the points stored
        by column.
    """

    def __init__(self):

        self._values = np.ndarray(shape=(0, 0))

    def append(self, point):
        """
        Add a new point.

        :param array_like point: the coordinates of the point to add.
        """
        array = np.asarray(point).reshape(-1, 1)
        try:
            self._values = np.append(self._values, array, 1)
        except ValueError:
            self._values = array

    def __getitem__(self, val):
        ret = Points()
        for point in self.values[:, val].T:
            ret.append(point)

        return ret

    @property
    def values(self):
        """
        The matrix that contains all the points, stored by column.

        :type: numpy.ndarray
        """
        return self._values

    @property
    def size(self):
        """
        The number of the points.

        :type: int
        """
        return self.values.shape[1]

    @property
    def dimension(self):
        """
        The dimension of the points.

        :type: int
        """
        return self.values.shape[0]

    @property
    def triangulation(self):
        """
        The Delaunay tasselation built from the points.

        :type: scipy.spatial.Delaunay
        """
        return Delaunay(self.values.T)



"""
Base module with the base class for reading and writing different files.
"""
import os
import ezyrb.stlhandler
import ezyrb.vtkhandler
import ezyrb.matlabhandler


class FileHandler(object):
    """
    A base class for file handling.
    """

    def __new__(cls, filename):
        """
        Generic file handler. When you create a new instance of FilaHandler, it
        returns specialized file handler (`MatlabHandler`, `VtkHandler`, ...)
        instance.

        :param str filename: name of file
        """
        if not isinstance(filename, str):
            raise TypeError("Filename must be a string")

        ext = {
            ".vtk": ezyrb.vtkhandler.VtkHandler,
            ".stl": ezyrb.stlhandler.StlHandler,
            ".mat": ezyrb.matlabhandler.MatlabHandler,
        }

        specialized_handler = ext.get(os.path.splitext(filename)[1])

        if specialized_handler is None:
            raise TypeError("Not supported file")

        return specialized_handler(filename)



"""
Derived module from filehandler.py to handle Vtk files.
"""
import os
import numpy as np
import vtk
import vtk.util.numpy_support as ns


class VtkHandler(object):
    """
    Vtk format file handler class.
    You are NOT supposed to call directly this class constructor (use
    :class:`.FileHandler` constructor instead)

    :param str filename: the name of the file to handle.

    :cvar str _filename: name of file to handle
    :cvar vtkPolyData _cached_data: private attribute to store the last polydata
        processed
    """

    def __init__(self, filename):

        self._filename = filename
        self._cached_data = None

    def _read_polydata(self):
        """
        This private method reads the given `filename` and return a vtkPolyData
        object containing all informations about file; to avoid useless IO
        operation on the same file, it stores polydata of the last file parsed
        and if user ask for this file, the polydata previously stored is
        returned.

        :return: polydata containing information about file.
        :rtype: vtkPolyData
        """
        # Polydata from `filename` is allready loaded; return it
        if self._cached_data is not None:
            return self._cached_data

        if not os.path.isfile(self._filename):
            raise RuntimeError("{0!s} doesn't exist".format(
                os.path.abspath(self._filename)))

        reader = vtk.vtkDataSetReader()
        reader.SetFileName(self._filename)
        reader.Update()
        data = reader.GetOutput()

        self._cached_data = data

        return data

    def _save_polydata(self, data, write_bin=False):
        """
        This private method saves into `filename` the `data`. `data` is a
        vtkPolydata. It is possible to specify format for `filename`: if
        `write_bin` is True, file is written in binary format, otherwise in
        ASCII format. This method save cached polydata to reduce number of IO
        operations.

        :param vtkPolyData data: polydatat to save.
        :param bool write_bin: for binary format file.
        """
        self._cached_data = data

        writer = vtk.vtkDataSetWriter()

        if write_bin:
            writer.SetFileTypeToBinary()

        writer.SetFileName(self._filename)

        if vtk.VTK_MAJOR_VERSION <= 5:
            writer.SetInput(data)
        else:
            writer.SetInputData(data)
        writer.Write()

    def get_geometry(self, get_cells=False):
        """
        This method reads the given filename and returns points and cells of
        file. If get_cells is True, it computes the list that contain index of
        points defining cells, otherwise the list is not computed and set to
        None (less expensive).

        :param bool get_cells: flag to compute cells list or not. Default is
            false.

        :return: the `n_points`-by-3 matrix containing the coordinates of the
            points, the `n_cells` list containing, for each cell, the iindex
            of the points that define the cell (if computed).
        :rtype: numpy.ndarray, list(numpy.ndarray)
        """
        data = self._read_polydata()

        n_points = data.GetNumberOfPoints()
        n_cells = data.GetNumberOfCells()

        points = np.array([data.GetPoint(i) for i in np.arange(n_points)])
        if get_cells:
            cells = [[
                data.GetCell(i).GetPointIds().GetId(idx)
                for idx in np.arange(data.GetCell(i).GetNumberOfPoints())
            ] for i in np.arange(n_cells)]
        else:
            cells = None

        return points, cells

    def set_geometry(self, points, cells, write_bin=False):
        """
        This method writes to `filename` a new data defined by `points` and
        `cells`.

        :param numpy.ndarray points: matrix *n_points*-by-3 containing
            coordinates of all points.
        :param list(array_like) cell: list that contains for each cell the list
            of the indices of points that define the cell.
        :param bool write_bin: flag to write in the binary format. Default is
            false.
        """
        data = vtk.vtkPolyData()
        vtk_points = vtk.vtkPoints()
        vtk_cells = vtk.vtkCellArray()

        for i in np.arange(points.shape[0]):
            vtk_points.InsertNextPoint(points[i])

        for i in np.arange(len(cells)):
            vtk_cells.InsertNextCell(len(cells[i]), cells[i])

        data.SetPoints(vtk_points)
        data.SetPolys(vtk_cells)

        self._save_polydata(data, write_bin)

    def get_dataset(self, output_name, datatype='point'):
        """
        This method reads the given `filename` and returns a numpy array
        containing `output_name` field for each point (if datatype is 'point')
        or for each cell (if datatype is 'cell').
        If `output_name` is not a field in file, it raises exception.

        :param str output_name: the name of the output of interest to extract
            from file.
        :param str datatype: a string to specify if point data or cell data
            should be returned. Default value is `point`.
        :return: the matrix `n_components`-by-`n_elements` containing the
            extracted output.
        :rtype: numpy.ndarray
        """
        if datatype not in ['cell', 'point']:
            raise ValueError("datatype MUST be 'cell' or 'point'")

        data = self._read_polydata()

        if datatype == 'point':
            extracted_data = data.GetPointData().GetArray(output_name)
        else:
            extracted_data = data.GetCellData().GetArray(output_name)

        if extracted_data is None:
            raise RuntimeError(
                datatype + " data has no " + output_name + " field.")

        output_values = ns.vtk_to_numpy(extracted_data)

        # Make 1D array a column array
        # (400, ) --> (400,1)
        try:
            cols = output_values.shape[1]
        except IndexError:
            cols = 1
        return output_values.reshape((-1, cols))

    def set_dataset(self,
                    output_values,
                    output_name,
                    datatype='point',
                    write_bin=False):
        """
        Writes to filename the given output. `output_values` is a matrix that
        contains the new values of output to write, `output_name` is a string
        that indicates name of output to write.

        :param numpy.ndarray output_values: the
            *n_points*-by-*n_components* matrix containing the output values.
        :param str output_name: name of the output.
        :param str datatype: a string to specify if point data or cell data
            should be returned. Default value is 'point'.
        :param bool write_bin: flag to write in the binary format. Default is
            false.
        """
        if datatype not in ['cell', 'point']:
            raise ValueError("datatype MUST be 'cell' or 'point'")

        data = self._read_polydata()

        output_array = ns.numpy_to_vtk(
            num_array=output_values, array_type=vtk.VTK_DOUBLE)
        output_array.SetName(output_name)

        if datatype == 'point':
            data.GetPointData().AddArray(output_array)
        else:
            data.GetCellData().AddArray(output_array)

        self._save_polydata(data, write_bin)

    def get_all_output_names(self):
        """
        Return the list of all the output names by point and the list of all the
        output names by cell.

        :return: point output name, cell output name
        :rtype: tuple(list(str), list(str))
        """

        data = self._read_polydata()

        point_name = [
            data.GetPointData().GetArrayName(i)
            for i in np.arange(data.GetPointData().GetNumberOfArrays())
        ]

        cell_name = [
            data.GetCellData().GetArrayName(i)
            for i in np.arange(data.GetCellData().GetNumberOfArrays())
        ]

        return point_name, cell_name



"""
Module for the generation of the reduced space by using Interpolation.
"""

from ezyrb.parametricspace import ParametricSpace
from scipy.interpolate import LinearNDInterpolator
import numpy as np


class Interpolation(ParametricSpace):
    """
    Documentation

    :cvar object _interpolator: interpolating object for the basis
        interpolation.
    """

    def __init__(self):

        self._interpolator = None

    @property
    def interpolator(self):
        """
        The multidimensional interpolator that combines the saved snapshots for
        the parametric space creation.

        :type: object
        """
        return self._interpolator

    def generate(self, points, snapshots):
        """
        Generate the reduced space using the proper orthogonal decomposition:
        the matrix that contains the `snapshots` computed for the `points` is
        decomposed using SVD tecnique to obtain the POD basis. The interpolator
        is built combining this basis.

        :param Snapshots snapshots: the snapshots.
        :param Points points: the parametric points where snapshots were
            computed.
        """
        self._interpolator = LinearNDInterpolator(points.values.T,
                                                  snapshots.values.T)

    def __call__(self, value):
        """
        Project a new parametric point onto the reduced space and a new
        approximated solution is provided.

        :param numpy.ndarray value: the new parametric point
        """
        return self._interpolator(value)

    @staticmethod
    def loo_error(points, snapshots, func=np.linalg.norm):
        """
        Compute the error for each parametric point with a leave-one-out (loo)
        strategy.

        :param Points points: the points where snapshots were computed.
        :param Snapshots snapshots: the snapshots.
        :param function func: the function to estimate error; default is the
            numpy.linalg.norm
        """
        loo_error = np.zeros(points.size)

        for j in np.arange(points.size):

            remaining_index = list(range(j)) + list(range(j + 1, points.size))
            remaining_snaps = snapshots[remaining_index]
            remaining_pts = points[remaining_index]

            subspace = Interpolation()
            subspace.generate(remaining_pts, remaining_snaps)
            projection = subspace(points[j].values)
            if projection is not float:
                projection = np.sum(remaining_snaps.values) / (points.size - 1)

            loo_error[j] = func(snapshots[j].values - projection)
            loo_error[j] /= func(snapshots[0].values)

        return loo_error



"""
Module for the manipulation of the solution
"""
from ezyrb.filehandler import FileHandler
import numpy as np


class Snapshots(object):
    """
    :param str output_name: the name of the output to extract from the solution
        files; it has to be the same for all the solution files.
    :param str weight_name: the name of the weights.
    :param str dformat: it allows to choose if the solutions are stored by
        points or by cells; allowed strings are 'cell' or 'point'. Default
        value is 'cell'.

    :cvar str _o_name: the output name.
    :cvar str _w_name: the weight name.
    :cvar numpy.ndarray _values: the solutions extracted from files, stored by
        column.
    :cvar numpy.ndarray _weighted: the weighted solutions extracted from files,
        stored by column.
    :cvar numpy.ndarray or int _weights: the weights extracted from a file.
    :cvar str _dformat: the flag to select the data type.
    :cvar list(str) _files: the files from which to extract the solutions.

    """

    def __init__(self, output_name, weight_name=None, dformat='cell'):

        self._o_name = output_name
        self._w_name = weight_name

        self._values = np.ndarray(shape=(0, 0))
        self._weighted = np.ndarray(shape=(0, 0))
        self._weights = 1.0

        self._dformat = dformat

        self._files = []

    def __getitem__(self, val):
        ret = Snapshots(self._o_name, self._w_name, self._dformat)
        ret._values = self._values[:, val]
        ret._weighted = self._weighted[:, val]
        ret._weights = self._weights
        if isinstance(val, list):
            ret._files = [self._files[i] for i in val]
        else:
            ret._files = self._files[val]

        return ret

    @property
    def files(self):
        """
        The list of file from which to extract the snapshots.

        :type: list(str)
        """
        return self._files

    @property
    def values(self):
        """
        The snapshots; it is a *n* x *m* matrix that contains the snapshots
        stored by column; *n* is the output dimension, *m* is the number of
        snapshots.

        :type: numpy.ndarray
        """
        return self._values

    @property
    def weighted(self):
        """
        The weighted snapshots; it is a *n* x *m* matrix that contains the
        snapshots stored by column; *n* is the output dimension, *m* is the
        number of snapshots.

        :type: numpy.ndarray
        """
        return self._weighted

    @property
    def weights(self):
        """
        The weights extracted.

        :type: int or array_like
        """
        return self._weights

    @property
    def size(self):
        """
        The number of the snapshots.

        :type: int
        """
        return self.values.shape[1]

    @property
    def dimension(self):
        """
        The dimension of the snapshots.

        :type: int
        """
        return self.values.shape[0]

    def append(self, filename):
        """
        Append the `filename` to the list of the solution files; update the
        values and the weighted values with the extracted solution.

        :param str filename: the name of file to append to the solution files.
        """
        if not isinstance(filename, str):
            raise TypeError

        if not self.values.size and self._w_name:

            self._weights = FileHandler(filename).get_dataset(
                self._w_name, self._dformat)

        self._files.append(filename)

        v = FileHandler(filename).get_dataset(self._o_name, self._dformat)

        self._append_values(v)
        self._append_weighted(v)

    def _append_values(self, values):
        """
        Private method to append an output.

        :param array_like values: the output.
        """

        array = np.asarray(values).reshape(-1, 1)
        try:
            self._values = np.append(self._values, array, 1)
        except ValueError:
            self._values = array

    def _append_weighted(self, values):
        """
        Private method to append a weighted output.

        :param array_like values: the output.
        """
        print(values.shape)
        array = np.dot(np.sqrt(self._weights), values.T).reshape(-1, 1)
        try:
            self._weighted = np.append(self._weighted, array, 1)
        except ValueError:
            self._weighted = array



"""
Module for the generation of the reduced space by using the Proper Orthogonal
Decomposition Interpolation
"""

from ezyrb.parametricspace import ParametricSpace
from scipy.interpolate import LinearNDInterpolator
import numpy as np


class PODInterpolation(ParametricSpace):
    """
    Documentation

    :cvar numpy.ndarray _pod_basis: basis extracted from the proper orthogonal
        decomposition.
    :cvar object _interpolator: interpolating object for the pod basis
        interpolation
    """

    def __init__(self):

        self._pod_basis = None
        self._interpolator = None

    @property
    def pod_basis(self):
        """
        The basis found by proper orthogonal decomposition.

        :type: numpy.ndarray
        """
        return self._pod_basis

    @property
    def interpolator(self):
        """
        The multidimensional interpolator that combines the saved snapshots for
        the parametric space creation.

        :type: object
        """
        return self._interpolator

    def generate(self, points, snapshots, interpolator=LinearNDInterpolator):
        """
        Generate the reduced space using the proper orthogonal decomposition
        interpolation: the matrix that contains the `snapshots` computed for
        the `points` is decomposed using SVD tecnique to obtain the POD basis.
        The POD basis are used to compute the modal coefficients. The
        interpolator combines this coefficients.

        :param Snapshots snapshots: the snapshots.
        :param Points points: the parametric points where snapshots were
            computed.
        :param object interpolator: the interpolator used to interpolate the
            coefficients.
        """
        eig_vec = np.linalg.svd(snapshots.weighted, full_matrices=False)[0]

        self._pod_basis = np.sqrt(snapshots.weights) * eig_vec
        coefs = self._pod_basis.T.dot(snapshots.weighted)
        self._interpolator = interpolator(points.values.T, coefs)

    def __call__(self, value):
        """
        Project a new parametric point onto the reduced space and a new
        approximated solution is provided.

        :param numpy.ndarray value: the new parametric point
        """
        return self._pod_basis.dot(self._interpolator(value).T)

    @staticmethod
    def loo_error(points, snapshots, func=np.linalg.norm):
        """
        Compute the error for each parametric point as projection of the
        snapshot onto the POD basis with a leave-one-out (loo) strategy.

        :param Points points: the points where snapshots were computed.
        :param Snapshots snapshots: the snapshots.
        :param function func: the function to estimate error; default is the
            :func:`numpy.linalg.norm`.
        """
        loo_error = np.zeros(points.size)

        for j in np.arange(points.size):

            remaining_index = list(range(j)) + list(range(j + 1, points.size))
            remaining_snaps = snapshots[remaining_index]

            eigvec = np.linalg.svd(
                remaining_snaps.weighted, full_matrices=False)[0]

            loo_basis = np.sqrt(remaining_snaps.weights) * eigvec

            projection = np.sum(
                np.array([
                    np.dot(snapshots[j].weighted, basis) * basis
                    for basis in loo_basis.T
                ]),
                axis=0)

            error = (snapshots[j].values - projection) * snapshots.weights
            loo_error[j] = func(error) / func(snapshots[0].values)

        return loo_error



"""
Module implementing some utilities for EZyRB
"""
import math
import numpy as np
from ezyrb.filehandler import FileHandler


def normal(p0, p1, p2):
    """
    Compute the surface normal of the surface indicated by three points.

    :param array_like p0: first point
    :param array_like p1: second point
    :param array_like p2: third point
    :return: the normal vector
    :rtype: numpy.ndarray
    """
    return np.cross(p1 - p0, p2 - p0)


def normalize(v):
    """
    Normalize a vector.

    :param array_like v: the vector to normalize
    :return: the normalized vector
    :rtype: numpy.ndarray
    """
    return v / np.linalg.norm(v, 2)


def polygon_area(points):
    """
    Compute the area of a planar non-self-intersecting polygon defined by
    `points`.

    :param numpy.ndarray points: a matrix that contains the vertices coordinates
        stored by row.
    :return: the area of the polygon
    :rtype: float
    """
    num_points = points.shape[0]

    if num_points < 3:
        return 0.0

    total = np.sum(
        [
            np.cross(points[i], points[(i + 1) % num_points])
            for i in np.arange(num_points)
        ],
        axis=0)

    unit_vector = normalize(normal(*points[0:3]))

    return np.abs(np.dot(total, unit_vector) * .5)


def simplex_volume(vertices):
    """
    Method implementing the computation of the volume of a N dimensional
    simplex.
    Source from: `wikipedia.org/wiki/Simplex
    <https://en.wikipedia.org/wiki/Simplex>`_.

    :param numpy.ndarray simplex_vertices: Nx3 array containing the
        parameter values representing the vertices of a simplex. N is the
        dimensionality of the parameters.

    :return: N dimensional volume of the simplex.
    :rtype: float
    """
    distance = np.transpose([vertices[0] - vi for vi in vertices[1:]])
    return np.abs(np.linalg.det(distance) / math.factorial(vertices.shape[1]))


def compute_area(filename):
    """
    Given a file, this method computes the area for each cell of the mesh stored
    in the file and returns it. It uses :func:`polygon_area`.

    :param str filename: the name of the file to parse in order to extract
        the necessary information about the cells.
    :return: the array that contains the area of each cells.
    :rtype: numpy.ndarray
    """
    points, cells = FileHandler(filename).get_geometry(get_cells=True)
    return np.array([polygon_area(points[cell]) for cell in cells])


def compute_normals(filename, datatype='cell'):
    """
    Given a file, this method computes the surface normals of the mesh stored
    in the file. It allows to compute the normals of the cells or of the points.
    The normal computed in a point is the interpolation of the cell normals of
    the cells adiacent to the point.

    :param str filename: the name of the file to parse in order to extract
        the geometry information.
    :param str datatype: indicate if the normals have to be computed for the
        points or the cells. The allowed values are: 'cell', 'point'. Default
        value is 'cell'.
    :return: the array that contains the normals.
    :rtype: numpy.ndarray
    """
    points, cells = FileHandler(filename).get_geometry(get_cells=True)
    normals = np.array(
        [normalize(normal(*points[cell][0:3])) for cell in cells])

    if datatype == 'point':
        normals_cell = np.empty((points.shape[0], 3))
        for i_point in np.arange(points.shape[0]):
            cell_adiacent = [cells.index(c) for c in cells if i_point in c]
            normals_cell[i_point] = normalize(
                np.mean(normals[cell_adiacent], axis=0))
        normals = normals_cell

    return normals


def write_area(filename, output_name='Area'):
    """
    Given a file, this method computes the area for each cell of the mesh stored
    in the file and save it as new dataset.

    :param str filename: the name of the file to parse in order to extract
        the geometry information.
    :param str output_name: the name of the new dataset that contains the
        cells area.
    """
    FileHandler(filename).set_dataset(
        compute_area(filename), output_name, datatype='cell')


def write_normals(filename, output_name='Normals', datatype='cell'):
    """
    Given a file, this method computes the surface normals of the mesh stored
    in the file and save it as new dataset.

    :param str filename: the name of the file to parse in order to extract
        the geometry information.
    :param str output_name: the name of the new dataset that contains the
        normals.
    :param str datatype: indicate if the normals have to be computed for the
        points or the cells. The allowed values are: 'cell', 'point'. Default
        value is 'cell'.
    """
    FileHandler(filename).set_dataset(
        compute_normals(filename), output_name, datatype=datatype)



"""
Module to handle STL files.
"""
import os
import numpy as np
import vtk


class StlHandler(object):
    """
    Vtk format file handler class.
    You are NOT supposed to call directly this class constructor (use
    :class:`~.FileHandler` constructor instead)

    :param str filename: name of file

    :cvar str _filename: name of file to handle
    :cvar vtkPolyData _cached_data: private attribute to store the last
        polydata processed
    """

    def __init__(self, filename):
        self._filename = filename
        self._cached_data = None

    def _read_polydata(self):
        """
        This private method reads the given `filename` and return a vtkPolyData
        object containing all informations about file; to avoid useless IO
        operation on the same file, it stores polydata of the last file parsed
        and if user ask for this file, the polydata previously stored is
        returned.

        :return: polydata containing information about file.
        :rtype: vtkPolyData
        """
        # Polydata from `filename` is already loaded; return it
        if self._cached_data is not None:
            return self._cached_polydata

        if not os.path.isfile(self._filename):
            raise RuntimeError("{0!s} doesn't exist".format(
                os.path.abspath(self._filename)))

        reader = vtk.vtkSTLReader()
        reader.SetFileName(self._filename)
        reader.Update()
        data = reader.GetOutput()

        self._cached_data = data

        return data

    def _save_polydata(self, data, write_bin=False):
        """
        This private method saves into `filename` the `data`. `data` is a
        vtkPolydata. It is possible to specify format for `filename`: if
        `write_bin` is True, file is written in binary format, otherwise in
        ASCII format. This method save cached polydata to reduce number of IO
        operations.

        :param vtkPolyData data: polydatat to save.
        :param bool write_bin: for binary format file.
        """
        self._cached_data = data

        writer = vtk.vtkSTLWriter()

        if write_bin:
            writer.SetFileTypeToBinary()

        writer.SetFileName(self._filename)

        if vtk.VTK_MAJOR_VERSION <= 5:
            writer.SetInput(data)
        else:
            writer.SetInputData(data)
        writer.Write()

    def get_geometry(self, get_cells=False):
        """
        This method reads the given `filename` and returns points and cells of
        file. If `get_cells` is True, it computes the list that contain index of
        points defining cells, otherwise the list is not computed and set to
        **None** (less expensive).

        :param bool get_cells: flag to compute cells list or not. Default is
            false.

        :return: the `n_points`-by-3 matrix containing the coordinates of the
            points, the n_cells list containing, for each cell, the index of
            the points that define the cell (if computed).
        :rtype: numpy.ndarray, list(numpy.ndarray) or None
        """
        data = self._read_polydata()

        n_points = data.GetNumberOfPoints()
        n_cells = data.GetNumberOfCells()

        points = np.array([data.GetPoint(i) for i in np.arange(n_points)])
        cells = None

        if get_cells:
            cells = [[
                data.GetCell(i).GetPointIds().GetId(idx)
                for idx in np.arange(data.GetCell(i).GetNumberOfPoints())
            ] for i in np.arange(n_cells)]

        return points, cells

    def set_geometry(self, points, cells, write_bin=False):
        """
        This method save into `filename` a new data defined by `points` and
        `cells`.

        :param numpy.ndarray points: matrix *n_points*-by-3 containing
            coordinates of all points.
        :param list(list(int)) cell: a list of lists that contains, for each
            cell, the index of points defining the cell.
        :param bool write_bin: flag to write in the binary format. Default is
            false.
        """
        data = vtk.vtkPolyData()
        vtk_points = vtk.vtkPoints()
        vtk_cells = vtk.vtkCellArray()

        for i in np.arange(points.shape[0]):
            vtk_points.InsertNextPoint(points[i])

        for i in np.arange(len(cells)):
            vtk_cells.InsertNextCell(len(cells[i]), cells[i])

        data.SetPoints(vtk_points)
        data.SetPolys(vtk_cells)

        self._save_polydata(data, write_bin)



from setuptools import setup, find_packages

def readme():
    """
    This function just return the content of README.md
    """
    with open('README.md') as f:
        return f.read()

setup(name='ezyrb',
      version='0.2',
      description='Easy Reduced Basis method',
      long_description=readme(),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
      ],
      keywords='dimension_reduction mathematics vtk pod podi',
      url='https://github.com/mathLab/EZyRB',
      author='Nicola Demo, Marco Tezzele',
      author_email='demo.nicola@gmail.com, marcotez@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
            'numpy',
            'scipy',
            'matplotlib',
            'enum34',
            'Sphinx>=1.4',
            'sphinx_rtd_theme',
            'yapf'
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
      zip_safe=False)



"""
Utilities to map solution on a specific geometry
"""
import sys
import numpy as np
from ezyrb.filehandler import FileHandler
import scipy


class Mapper(object):
    """
    Documentation

    :cvar list(str) _output_name: the list of output names to map.
    :cvar str _mode: indicate if new file will be interpolated using point
        data or cell data.
    :cvar int _n_neighbors: number of neighbors to use to interpolate new
        value.
    :cvar function _interpolate_func: function to interpolate new value
        starting from neighbors value.
    """

    def __init__(self):
        self._output_name = []
        self._neighbour_locator = None
        self._mode = "point"
        self._n_neighbors = 1
        self.__interpolate_func = Mapper.default_interpolate_func

    @staticmethod
    def get_cell_centroid(vertices):
        """
        Compute the centroid of the cell defined by `vertices`; the centroid is
        computed as the `vertices` coordinates average.

        :param numpy.ndarray vertices: a *n_vertices*-by-3 matrix where vertices
            coordinates are store by row.
        :return: cell centroid coordinates
        :rtype: numpy.ndarray
        """
        centroid = np.mean(vertices, axis=0)
        return centroid

    @staticmethod
    def default_interpolate_func(values, distance):
        """
        Default function for the point interpolation. Take as argument
        the neighbors values and the respective distance between
        to-interpolate point and neighbors, and return the values weighted
        average (weights are the reciprocal of the distances). If
        to-interpolate point coincides with one of the neighbors, return that
        neighbour value.

        :param numpy.ndarray values: a matrix *number_of_neighbors* -by-
            *number_of_components* containing neighbors values.
        :param numpy.narray distance: a vector containing distance between
            to-interpolate point and neighbors.
        """
        if not distance.all():
            return values[np.where(distance == 0)]

        return np.average(values, weights=np.reciprocal(distance), axis=0)

    @property
    def output_name(self):
        """
        The names of outputs you wanna interpolate from the original solution
        file to the mapped one.

        :getter: Returns the output names
        :setter: Sets the output names
        :type: list(str)
        """
        return self._output_name

    @output_name.setter
    def output_name(self, output_name):

        if not isinstance(output_name, list):
            output_name = [output_name]

        for name in output_name:
            if not isinstance(name, str):
                raise TypeError

        self._output_name = output_name

    @property
    def interpolate_function(self):
        """
        The function to interpolate value of a new point using neighbors value.
        It has to take as arguments two `numpy.array`: the first indicates
        neighbors value (one row for each value) and the second indicates
        distance from interpolated point.

        :getter: gets the interpolated function
        :setter: sets the interpolation function
        :type: function
        """
        return self.__interpolate_func

    @interpolate_function.setter
    def interpolate_function(self, func):

        if not callable(func):
            raise TypeError

        self.__interpolate_func = func

    @property
    def interpolation_mode(self):
        """
        Indicate if solution will be mapped using the selected output point data
        or the selected output cell data. Valid value is 'point' or 'cell'.

        :getter: gets current mode
        :setter: sets the mode
        :type: string
        """
        return self._mode

    @interpolation_mode.setter
    def interpolation_mode(self, mode):

        if not isinstance(mode, str):
            raise TypeError

        if mode not in ['cell', 'point']:
            raise ValueError

        self._mode = mode

    def _build_neighbour_locator(self, points):
        """
        Construct kdtree to nearest neighbors search.

        :param numpy.ndarray points: a *n_points* -by- 3 where are stored
            coordinates of points by row
        """
        try:
            self._neighbour_locator = scipy.spatial.cKDTree(points)
        except RuntimeError:
            # The maximum recursion limit can be exceeded for large data sets.
            sys.setrecursion(points.shape[0])
            self._neighbour_locator = scipy.spatial.cKDTree(points)

    def _find_neighbour(self, coordinates):
        """
        This method looks for the nearest neighbors to a given point; the
        locator have to build before looking for operation.

        :param numpy.array coordinates: coordinates of the point to query
        :return: a 2 *n_neighbors*-by-*n_query_points*; it contains in the
            first `number_neighbors` columns the index of neighbors , and in
            last *number_neighbors* columns the distance from query points to
            neighbors.
        :rtype: numpy.ndarray
        """
        if self._neighbour_locator is None:
            raise RuntimeError("Neighbour locator seems not initialized")

        distance, id_neighbour = self._neighbour_locator.query(
            coordinates, self._n_neighbors)

        return np.concatenate(
            (np.array(id_neighbour).reshape(
                (-1, self._n_neighbors)), np.array(distance).reshape(
                    (-1, self._n_neighbors))),
            axis=1)

    @property
    def number_neighbors(self):
        """
        The number of neighbors to use for new value interpolation.

        :getter: Sets number of neighbors
        :setter: Return number of neighbors
        :type: int
        """
        return self._n_neighbors

    @number_neighbors.setter
    def number_neighbors(self, n):

        if not isinstance(n, int):
            raise TypeError

        self._n_neighbors = n

    def map_solution(self, output_file, solution_file, geometry_file=None):
        """
        This method maps the solution on a new geometry. To be more
        specific, for each point (or cell) of the given geometry looks for the
        nearest solution points (or cells) and interpolate the selected output
        using neighbors values.It is possible select number of neighbors and the
        interpolate function. If geometry_file is not specified, to-interpolate
        points are taken from output_file, so in this case output_file must
        contain geometry information; if geometry_file is specified,
        to-interpolate points are taken from geometry_file and the interpolated
        solution is write on output_file.

        :param str output_file: name of file where interpolated solution will
            be stored. Only vtk file are supported.
        :param str solution_file: name of file where real solution is stored.
        :param str geometry_file: name of file where geometry information is
            stored.
        """

        if not geometry_file:
            geometry_file = output_file

        geo_file_handler = FileHandler(geometry_file)
        sol_file_handler = FileHandler(solution_file)
        out_file_handler = FileHandler(output_file)

        #####
        # Read points and cell from geometry file and solution file
        #
        point_sol, cell_sol = sol_file_handler.get_geometry(get_cells=True)
        point_map, cell_map = geo_file_handler.get_geometry(get_cells=True)

        #####
        # Only if geomtry file is not output file, I have to copy geomtry
        # to output file
        if geometry_file != output_file:
            out_file_handler.set_geometry(point_map, cell_map)

        #####
        # Only for 'cell' mapping
        #
        # if mapping using cell, compute cells centroid and use them
        # as dataset for neighbors locator to find closest cell
        if self.interpolation_mode == "cell":
            centroid_sol = np.array(
                [self.get_cell_centroid(point_sol[cell]) for cell in cell_sol])
            centroid_map = np.array(
                [self.get_cell_centroid(point_map[cell]) for cell in cell_map])

            point_sol = centroid_sol
            point_map = centroid_map

        #####
        # Build locator using correct points
        self._build_neighbour_locator(point_sol)

        #####
        # For each point find closest points in solution geometry
        neighbors = np.array([
            self._find_neighbour(query_point).ravel()
            for query_point in point_map
        ])

        #####
        # Here the real mapping of new points on solution points
        for output_name in self._output_name:
            output_array = sol_file_handler.get_dataset(
                output_name, datatype=self._mode)

            index_near = neighbors[:, 0:self.number_neighbors].astype(int)
            distance = neighbors[:, self.number_neighbors:]

            mapped_output = np.array([
                self.interpolate_function(output_array[index_near[id_neigh]],
                                          distance[id_neigh])
                for id_neigh in np.arange(neighbors.shape[0])
            ])

            out_file_handler.set_dataset(
                mapped_output, output_name, datatype=self._mode)



"""
Utilities for the online evaluation of the output of interest
"""
from ezyrb.filehandler import FileHandler
from ezyrb.parametricspace import ParametricSpace


class Online(object):
    """
    Online phase

    :param str output_name: the name of the output of interest.
    :param str space_filename: the name of the file where the space has been
        saved.
    :param str dformat: the data format to use for save the output to new file:
        if the parameter is "cell", the approximated output will be saved to
        cell data, if the parameter is "point", the approximated output will be
        saved to the point data.  These are the only options available. Default
        is 'cell'.

    :cvar str output_name: the name of the output of interest.
    :cvar ParametricSpace space_type: the type of space used for the online
        phase.
    :cvar str dformat: the data format to use for save the output to new file:
        if the parameter is "cell", the approximated output will be saved to
        cell data, if the parameter is "point", the approximated output will be
        saved to the point data.  These are the only options available. Default
        is 'cell'.
    """

    def __init__(self, output_name, space_filename, dformat='cell'):

        self.output_name = output_name
        self.dformat = dformat
        self.space = ParametricSpace.load(space_filename)

    def run(self, value):
        """
        This method evaluates the new point `value` in the parametric space and
        returns the approximated solution.

        :param array_like value: the point where the approximated solution has
            to be evaluated.

        :return: the approximated solution.
        :rtype: numpy.ndarray
        """
        return self.space(value)

    def run_and_store(self, value, filename, geometry_file=None):
        """
        This method evaluates the new point `value` in the parametric space and
        save the approximated solution on `filename`. It is possible to pass as
        optional argument the `geometry_file` that contains the topology on
        which the solution is projected.

        :param array_like value: the point where the approximated solution has
            to be evaluated.
        :param str filename: the file where the approximated solution is
            projected.
        :param str geometry_filename: the file that contains the topology to
            use for the solution projection.
        """
        output = self.space(value)
        writer = FileHandler(filename)
        if geometry_file:
            points, cells = FileHandler(geometry_file).get_geometry(True)
            writer.set_geometry(points, cells)

        writer.set_dataset(output, self.output_name, datatype=self.dformat)
